/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000, 2001  Pan Development Team <pan@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <limits.h>
#include <string.h>
#include <stdlib.h>

#include <glib.h>

#include <pan/base/acache.h>
#include <pan/base/article-thread.h>
#include <pan/base/debug.h>
#include <pan/base/file-headers.h>
#include <pan/base/group.h>
#include <pan/base/log.h>
#include <pan/base/pan-i18n.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/newsrc.h>
#include <pan/base/pan-callback.h>
#include <pan/base/server.h>

static void fire_groups_changed (Group **, gint qty);

/**
***
**/

Group*
group_new (Server * server, const char * name)
{
	Group * group;
	
	g_return_val_if_fail (server_is_valid (server), NULL);
	pan_warn_if_fail (is_nonempty_string(name));

	group = server_alloc_new_group (server);
	group_constructor (group, name);
	return group;
}

void
group_constructor (Group                * g,
		   const char           * name)
{
	pan_warn_if_fail (g!=NULL);
	pan_warn_if_fail (is_nonempty_string(name));

	debug1 (DEBUG_PAN_OBJECT, "group_constructor: %p", g);
	g->article_mutex = g_mutex_new ();
	g->loaded_since_last_fetch = FALSE;
	g->articles_dirty = FALSE;
	g->flags = (gint8)0;
	g->old_sort_style = (gint8)0;
	g->new_sort_style = (gint8)0;
	g->filter_bits = ~((guint16)STATE_FILTER_IGNORED);
	g->filter_show = FILTER_SHOW_MATCHES_AND_REFERENCES;
	g->article_qty = (gint32)0;
	g->article_read_qty = (gint32)0;
	g->article_low = (gulong)0;
	g->article_high_old = (gulong)0;
	g->permission = '?';
	g->article_high = (gulong)0;
	g->name = g_strdup (name);
	g->filter_name = NULL;
	g->identity_name = NULL;
	g->description = NULL;
	g->readable_name = NULL;
	g->download_dir = NULL;
	g->chunk = NULL;

	g->_articles_refcount = (gint)0;
	g->_article_chunk = NULL;
	g->_newsrc = NULL;
	g->_purged = NULL;
	g->_articles = NULL;
	g->_deleted_articles = NULL;
	g->_one_big_chunk = NULL;
}

static void
destroy_articles_ghfunc (gpointer key, gpointer val, gpointer user_data)
{
	article_destructor (ARTICLE(val));
}

void
group_destructor (Group * o)
{
	Group * g = GROUP(o);

	g_return_if_fail (o!=NULL);

	debug1 (DEBUG_PAN_OBJECT, "group_destructor: %p", g);

	/* strings */
	replace_gstr (&g->name, NULL);
	replace_gstr (&g->description, NULL);
	replace_gstr (&g->readable_name, NULL);
	replace_gstr (&g->download_dir, NULL);
	replace_gstr (&g->filter_name, NULL);
	replace_gstr (&g->identity_name, NULL);

	/* newsrc */
	if (g->_newsrc != NULL)
		pan_object_unref (PAN_OBJECT(g->_newsrc));
	if (g->_purged != NULL)
		pan_object_unref (PAN_OBJECT(g->_purged));

	/* articles */
	pan_warn_if_fail (g->_articles_refcount >= 0);
	replace_gstr (&g->_one_big_chunk, NULL);
	if (g->_articles != NULL) {
		g_hash_table_foreach (g->_articles, destroy_articles_ghfunc, NULL);
		g_hash_table_destroy (g->_articles);
		g->_articles = NULL;
	}
	if (g->_deleted_articles != NULL) {
		pan_g_ptr_array_foreach (g->_deleted_articles, (GFunc)article_destructor, NULL);
		g_ptr_array_free (g->_deleted_articles, TRUE);
		g->_deleted_articles = NULL;
	}
	if (g->chunk != NULL) {
		g_string_chunk_free (g->chunk);
		g->chunk = NULL;
	}
	if (g->_article_chunk != NULL) {
		memchunk_destroy (g->_article_chunk);
		g->_article_chunk = NULL;
	}
	g_mutex_free (g->article_mutex);
}

/***
****
****  PUBLIC ACCESSORS / MUTATORS
****
***/

void
group_set_new (Group * group, gboolean is_new)
{
	g_return_if_fail (group != NULL);

	if (is_new)
		group_set_flags (group, group->flags|GROUP_NEW);
	else
		group_set_flags (group, group->flags&~GROUP_NEW);
}

gboolean 
group_is_new (const Group * group)
{
	g_return_val_if_fail (group_is_valid(group), FALSE);

	return (group->flags & GROUP_NEW) ? 1 : 0;
}

void
groups_set_subscribed (Group ** groups, gint qty, gboolean subscribed)
{
	gint i;
	ServerGroupsType type;
	Server * server = NULL;

	/* sanity clause */
	g_return_if_fail (groups != NULL);
	g_return_if_fail (qty > 0);

	type = SERVER_GROUPS_SUBSCRIBED|SERVER_GROUPS_UNSUBSCRIBED;

	for (i=0; i<qty; ++i)
	{
		Group * group = groups[i];

		if (group->server != NULL)
		{
			server = group->server;
			server_set_group_type_dirty (server, type);
		}

		if (subscribed)
			group_set_flags (group, group->flags|GROUP_SUBSCRIBED);
		else
			group_set_flags (group, group->flags&~GROUP_SUBSCRIBED);
	}

	if (server != NULL)
		server_save_grouplist_if_dirty (server, NULL);

	fire_groups_changed (groups, qty);
}

gboolean 
group_is_subscribed (const Group * group)
{
	g_return_val_if_fail (group_is_valid(group), FALSE);

	return (group->flags & GROUP_SUBSCRIBED) ? 1 : 0;
}

void
group_set_dirty (Group * group)
{
	g_return_if_fail (group_is_valid(group));

	if (group->server != NULL)
		server_set_group_type_dirty (group->server, server_get_group_type (group));
}

void
group_set_download_dir (Group * group, const gchar* download_dir)
{
	g_return_if_fail (group_is_valid(group));

	replace_gstr (&group->download_dir, g_strdup(download_dir));
}

void
group_set_flags (Group * group, guint flags)
{
	g_return_if_fail (group_is_valid(group));

	group->flags = flags;
	group_set_dirty (group);
}

void
group_set_filter (Group         * group,
                  guint           filter_bits,
                  gulong          filter_show,
                  const gchar   * filter_name)
{
	g_return_if_fail (group_is_valid(group));

	group->filter_bits = filter_bits;
	group->filter_show = filter_show;
	replace_gstr (&group->filter_name, g_strdup(filter_name));

	group_set_dirty (group);
}

void
group_set_identity (Group       * group,
                    const gchar * identity_name)
{
	g_return_if_fail (group!=NULL);

	replace_gstr (&group->identity_name, g_strdup (identity_name));

	group_set_dirty (group);
}

void
group_set_sort_style (Group * group, int sort_style)
{
	g_return_if_fail (group != NULL);

	if (abs(sort_style) != abs(group->new_sort_style))
		group->old_sort_style = group->new_sort_style;
	group->new_sort_style = sort_style;
	group_set_dirty (group);
}

void
group_set_article_range (Group    * g,
                         gulong     article_low,
                         gulong     article_high)
{
	g_return_if_fail (group_is_valid(g));

	g->article_low = article_low;
	g->article_high = article_high;
	newsrc_set_group_range (group_get_newsrc(g), g->article_low, g->article_high);
	newsrc_set_group_range (group_get_purged(g), g->article_low, g->article_high);

	group_set_dirty (g);
}

void
group_get_article_range (const Group   * g,
                         gulong        * article_low,
                         gulong        * article_high)
{
	g_return_if_fail (group_is_valid(g));
	g_return_if_fail (article_low!=NULL);
	g_return_if_fail (article_high!=NULL);

	*article_low = g->article_low;
	*article_high = g->article_high;
}


void
group_mark_new_article_number (Group         * group,
                               gulong          everything_above_this_is_new)
{
	g_return_if_fail (group_is_valid(group));

	group->article_high_old = everything_above_this_is_new;
}

void
group_set_article_qty (Group * group, gint article_qty)
{
	g_return_if_fail (group_is_valid(group));

	if (article_qty != group->article_qty)
	{
		group->article_qty = article_qty;
		group_set_dirty (group);
		fire_groups_changed (&group, 1);
	}
}

void
group_set_article_read_qty (Group * group, gint article_read_qty)
{
	g_return_if_fail (group_is_valid(group));

	article_read_qty = MAX (article_read_qty, 0);
	group->article_read_qty = article_read_qty;
	group_set_dirty (group);
	fire_groups_changed (&group, 1);
}
void
group_inc_article_read_qty (Group * group, gint inc)
{
	g_return_if_fail (group_is_valid(group));

	if (inc != 0)
		group_set_article_read_qty (group, group->article_read_qty + inc);
}

void
group_mark_all_read (Group * g, gboolean read)
{
	Newsrc * newsrc;
	GPtrArray * articles;
	debug_enter ("group_mark_all_read");

	/* sanity clause */
	g_return_if_fail (group_is_valid(g));

	/* ensure the articles are loaded */
	group_ref_articles (g, NULL);

	/* mark all the articles as read.  use articles_set_read for this 
	   because that updates all the unread_children fields, marks 
	   crossposts as read, etc... */

	g_mutex_lock (g->article_mutex);
	articles = group_get_article_array (g);
	if (articles->len > 0)
		articles_set_read  ((Article**)articles->pdata, articles->len, TRUE);
	g_ptr_array_free (articles, TRUE);
	g_mutex_unlock (g->article_mutex);

	/**
	 * Wipe the entire newsgroup's article range as read.
	 * This reduces the group's newsrc to a single range "low-high"
	 * instead of thousands of smaller ranges.
	 */
	newsrc = group_get_newsrc (g);
	newsrc_set_group_range (newsrc, g->article_low, g->article_high);
	newsrc_mark_all (newsrc, read);

	/* update group counts */
	group_set_article_read_qty (g, (read ? g->article_qty : 0));

	/* unref the articles */
	group_unref_articles (g, NULL);

	debug_exit ("group_mark_all_read");
}

gboolean
group_is_read_only (const Group * group)
{
	g_return_val_if_fail (group_is_valid(group), FALSE);

	return group->permission == 'n';
}

gboolean
group_is_moderated (const Group * group)
{
	g_return_val_if_fail (group_is_valid(group), FALSE);

	return group->permission == 'm';
}

gboolean
group_is_valid (const Group * group)
{
	g_return_val_if_fail (group!=NULL, FALSE);
	g_return_val_if_fail (server_is_valid (group->server), FALSE);
	g_return_val_if_fail (group->article_mutex!=NULL, FALSE);
	g_return_val_if_fail (is_nonempty_string(group->name), FALSE);

	return TRUE;
}

gboolean
group_is_group (const Group * group)
{
	g_return_val_if_fail (group_is_valid(group), FALSE);

	return group->flags & GROUP_FOLDERS ? 0 : 1;
}

gboolean
group_is_folder (const Group * group)
{
	g_return_val_if_fail (group_is_valid(group), FALSE);

	return group->flags & GROUP_FOLDERS ? 1 : 0;
}

void
group_set_is_folder (Group * group, gboolean folder)
{
	g_return_if_fail (group_is_valid(group));

	if (!folder != !group_is_folder(group))
	{
		if (folder)
			group->flags |= GROUP_FOLDERS;
		else
			group->flags &= ~GROUP_FOLDERS;
		group_set_dirty (group);
	}
}

/***
****
****  DELETE GROUP
****
***/

void
group_empty (Group * group, gboolean clear_counts)
{
	GPtrArray * a;

	/* sanity clause */
	g_return_if_fail (group_is_valid(group));

	/* remove the articles from disk */
	file_headers_destroy (group);

	/* remove the articles from memory */
	a = group_get_article_array (group);
	group_remove_articles (group, a);
	g_ptr_array_free (a, TRUE);

	/* reset the counters */
	group_set_article_qty (group, 0);
	group_set_article_read_qty (group, 0);

	/* if user wants to clear the range, do that too */
	if (clear_counts)
		group_set_article_range (group, 0, 0);
}

/**
***  ARTICLE LIST
**/

static void
group_add_articles_ghfunc (gpointer key, gpointer val, gpointer data)
{
	Group * group = GROUP(key);
	GPtrArray * articles = (GPtrArray*)val;

	/* remove the articles from the other group */
	group_remove_articles (group, articles);

	/* clean up */
	g_ptr_array_free (articles, TRUE);
}

void
group_add_article (Group * group, Article * article)
{
	GPtrArray * a;
	debug_enter ("group_add_article");

	a = g_ptr_array_new ();
	g_ptr_array_add (a, article);
	group_add_articles (group, a, NULL, NULL, NULL);
	g_ptr_array_free (a, TRUE);

	debug_exit ("group_add_article");
}

static void
group_add_articles_impl (Group             * group,
                         const GPtrArray   * articles,
                         StatusItem        * status,
                         GPtrArray         * used,
                         GPtrArray         * ignored,
                         gboolean            lock_group)
{
	gint old_qty;
	gint new_read_qty;
	guint i;
	gulong low;
	gulong high;
	GPtrArray * tmp_used;
	GHashTable * ours;
	Newsrc * newsrc;
	Newsrc * purged;
	debug_enter ("group_add_articles");

	/* sanity checks */
	g_return_if_fail (group_is_valid(group));
	g_return_if_fail (group->_articles_refcount>=0);
	g_return_if_fail (articles!=NULL);
	if (!articles->len) return;
	g_return_if_fail (articles->len>0);
	g_return_if_fail (articles_are_valid ((const Article**)articles->pdata, articles->len));

	/**
	***  Remove the articles from their old groups.
	***  Batching the removal makes the code uglier but
	***  it runs faster.
	**/

	if (1)
	{
		GHashTable * remove_from_old = NULL;

		for (i=0; i!=articles->len; ++i)
		{
			Article * article = ARTICLE(g_ptr_array_index(articles,i));
			Group * oldgroup = article->group;
			gboolean array_existed;
			GPtrArray * to_remove;

			if (oldgroup==NULL || oldgroup==group)
				continue;

			/* ensure the hash table exists */
			if (remove_from_old == NULL)
				remove_from_old = g_hash_table_new (g_direct_hash, g_direct_equal);

			/* add this array to the array of articles to remove from its group */
			to_remove = g_hash_table_lookup (remove_from_old, oldgroup);
			array_existed = to_remove != NULL;
			if (!array_existed)
				to_remove = g_ptr_array_new ();
			g_ptr_array_add (to_remove, article);

			/* we hash the 'remove these' array by group */
			if (!array_existed)
				g_hash_table_insert (remove_from_old, oldgroup, to_remove);
		}

		if (remove_from_old != NULL)
		{
			g_hash_table_foreach (remove_from_old, group_add_articles_ghfunc, NULL);
			g_hash_table_destroy (remove_from_old);
		}
	}

	/**
	***  Get a handle on our current stats...
	**/

	low = group->article_low;
	high = group->article_high;

	/**
	***  Now try to add the articles that the client passed in
	**/

	new_read_qty = 0;
	tmp_used = g_ptr_array_new ();
	newsrc = group_get_newsrc (group);
	purged = group_get_purged (group);
	pan_g_ptr_array_reserve (tmp_used, articles->len);

	if (lock_group) {
		g_mutex_lock (group->article_mutex);
	}

	ours = group_get_articles (group);
	old_qty = (int) g_hash_table_size (ours);
	for (i=0; i!=articles->len; ++i)
	{
		Article * a = ARTICLE(g_ptr_array_index(articles,i));
		const gchar * message_id = article_get_message_id (a);
		gboolean is_purged = newsrc_is_article_read(group->_purged, a->number);

		if (!is_purged && g_hash_table_lookup(ours,message_id) == NULL)
		{
			a->group = group;
			g_hash_table_insert (ours, (gpointer)message_id, a);
			g_ptr_array_add (tmp_used, a);

			if (a->number<low)
				low  = a->number;
			if (a->number>high)
				high = a->number;
			if (newsrc_is_article_read (newsrc, a->number))
				++new_read_qty;
		}
		else if (ignored != NULL)
		{
			g_ptr_array_add (ignored, a);
		}
	}

	/**
	***  Rethread the articles
	**/

	if (tmp_used->len != 0) {
		GPtrArray * a = group_get_article_array (group);
		thread_articles (a);
		g_ptr_array_free (a, TRUE);
	}

	/**
	***  Unlock the article mutex
	**/

	if (lock_group) {
		g_mutex_unlock (group->article_mutex);
	}

	/**
	***  Now update our stats if articles were added
	**/

	if (tmp_used->len != 0)
	{
		/* maybe update the article range */
		if (high!=group->article_high || low!=group->article_low)
			group_set_article_range (group, low, high);

		/* update the group article stats */
		group_set_article_qty (group, g_hash_table_size(ours));
		if (old_qty != 0)
			new_read_qty += group->article_read_qty;
		group_set_article_read_qty (group, new_read_qty);
		group_set_articles_dirty (group);
		fire_groups_changed (&group, 1);

		pan_callback_call (group_get_articles_added_callback(), group, tmp_used);
	}

	if (used != NULL)
		pan_g_ptr_array_assign (used, tmp_used->pdata, tmp_used->len);

	/* cleanup */
	g_ptr_array_free (tmp_used, TRUE);
		
	debug_exit ("group_add_articles");
}

static void
group_add_articles_remove_unused_impl (Group            * group,
                                       GPtrArray        * articles,
                                       StatusItem       * status,
				       gboolean           lock)
{
	GPtrArray * ignored;
	GPtrArray * used;

	/* entry assertions */
	g_return_if_fail (group_is_valid(group));
	g_return_if_fail (articles != NULL);
	g_return_if_fail (articles->len != 0);

	/* add the articles */
	ignored  = g_ptr_array_new ();
	used = g_ptr_array_new ();
	group_add_articles_impl (group, articles, status, used, ignored, lock);

	/* destroy unused articles & remove from the incoming array */
	pan_g_ptr_array_assign (articles, used->pdata, used->len);
	pan_g_ptr_array_foreach (ignored, (GFunc)article_destructor, NULL);

	/* cleanup */
	g_ptr_array_free (ignored, TRUE);
	g_ptr_array_free (used, TRUE);
}

void
group_add_articles_remove_unused (Group            * group,
                                  GPtrArray        * articles,
                                  StatusItem       * status)
{
	group_add_articles_remove_unused_impl (group, articles, status, TRUE);
}

void
group_init_articles (Group * group,
                     GPtrArray * articles,
                     StatusItem * status)
{
	group_add_articles_remove_unused_impl (group, articles, status, FALSE);
}

void
group_add_articles (Group             * group,
                    const GPtrArray   * articles,
                    StatusItem        * status,
                    GPtrArray         * used,
                    GPtrArray         * ignored)
{
	group_add_articles_impl (group, articles, status, used, ignored, TRUE);
}

void
group_set_articles_dirty (Group * group)
{
	g_return_if_fail (group!=NULL);

	group->articles_dirty = TRUE;
}

gpointer
group_get_article_by_message_id (Group * group, const gchar * message_id)
{
	g_return_val_if_fail (group_is_valid(group), NULL);
	g_return_val_if_fail (is_nonempty_string(message_id), NULL);
	g_return_val_if_fail (*message_id=='<', NULL);

	return g_hash_table_lookup (group_get_articles(group), message_id);
}

void
group_remove_article (Group * group, Article * article)
{
	GPtrArray * a;

	g_return_if_fail (group_is_valid(group));
	g_return_if_fail (article_is_valid(article));

       	a = g_ptr_array_new ();
	g_ptr_array_add (a, article);
	group_remove_articles (group, a);
	g_ptr_array_free (a, TRUE);
}

void
group_remove_articles (Group * group, const GPtrArray * articles)
{
	guint i;
	int removed_read;
	GPtrArray * tmp_remove;
	GPtrArray * removed;
	GHashTable * ours;
	debug_enter ("group_remove_articles");

	g_return_if_fail (group_is_valid(group));

	/* sanity clause */
	g_return_if_fail (group != NULL);
	g_return_if_fail (articles != NULL);
	if (!articles->len) return;
	for (i=0; i!=articles->len; ++i)
		g_return_if_fail (ARTICLE(g_ptr_array_index(articles,i))->group==group);

	/* let everyone know that articles are going to be removed */
	tmp_remove = pan_g_ptr_array_dup ((GPtrArray*)articles);
	sort_articles ((Article**)tmp_remove->pdata, tmp_remove->len, ARTICLE_SORT_MSG_ID, TRUE);
	pan_callback_call (group_get_articles_removed_callback(), group, tmp_remove);

	/**
	***  Try to remove the articles
	**/

	g_mutex_lock (group->article_mutex);
	ours = group_get_articles (group);
	removed_read = 0;
	removed = g_ptr_array_new ();
	for (i=0; i!=tmp_remove->len; ++i)
	{
		Article * a = ARTICLE(g_ptr_array_index(tmp_remove,i));
		const gchar * message_id = article_get_message_id (a);
		Article * our_a = ARTICLE(g_hash_table_lookup (ours, message_id));

		if (our_a != NULL)
		{
			g_hash_table_remove (ours, message_id);

			/* add to the 'removed' array */
			g_ptr_array_add (removed, our_a);

			/* count how many removed articles were read */
			if (article_is_read (our_a))
				++removed_read;
		}
	}

	/**
	***  Rethread the affected articles
	**/

	if (removed->len)
	{
		GPtrArray * rethread_a = article_get_unique_threads (removed, GET_WHOLE_THREAD);
		GHashTable * rethread_hash = g_hash_table_new (g_str_hash, g_str_equal);

		/* add removed articles to the 'deleted' list */
		if (group->_deleted_articles == NULL)
			group->_deleted_articles = g_ptr_array_new ();
		pan_g_ptr_array_append (group->_deleted_articles, removed->pdata, removed->len);

		/* build a hash table of the articles that need to be rethreaded */
		for (i=0; i!=rethread_a->len; ++i) {
			Article * a = ARTICLE(g_ptr_array_index(rethread_a, i));
			g_hash_table_insert (rethread_hash, (gpointer)a->message_id, a);
		}
		/* remove the deleted articles from this rethread list */
		for (i=0; i!=removed->len; ++i) {
			Article * a = ARTICLE(g_ptr_array_index(removed, i));
			g_hash_table_remove (rethread_hash, a->message_id);
		}

		/* rethread the affected articles */
		if (g_hash_table_size(rethread_hash) > 0) {
			pan_hash_to_ptr_array (rethread_hash, rethread_a);
			thread_articles (rethread_a);
		}

		/* cleanup */
		g_hash_table_destroy (rethread_hash);
		g_ptr_array_free (rethread_a, TRUE);
	}

	/**
	***  Unlock the lock
	**/

	g_mutex_unlock (group->article_mutex);

	/**
	***  Update the stats
	**/

	if (removed->len != 0)
	{
		/* update the articles stats */
		group->article_qty -= removed->len;
		group->article_read_qty -= removed_read;
		group_set_articles_dirty (group);
		fire_groups_changed (&group, 1);
	}

	/* cleanup */
	g_ptr_array_free (removed, TRUE);
	g_ptr_array_free (tmp_remove, TRUE);
	debug_exit ("group_remove_articles");
}

void
group_expire_articles_not_in_range (Group * g, gulong low, gulong high)
{
	guint i;
	GPtrArray * tmp;
	GPtrArray * removeme;
	debug_enter ("group_expire_articles_not_in_range");

	/* sanity clause */
	g_return_if_fail (group_is_valid(g));

	/* find old articles to remove */
	tmp = group_get_article_array (g);
	removeme = g_ptr_array_new ();
	pan_g_ptr_array_reserve (removeme, tmp->len);
	for (i=0; i!=tmp->len; ++i) {
		Article * a = ARTICLE(g_ptr_array_index(tmp,i));
		if ((a->number<low) || (a->number>high))
			g_ptr_array_add (removeme, a);
	}

	/* remove the old articles */
	if (removeme->len != 0)
	{
		log_add_va (LOG_INFO, _("Expired %u articles from `%s'"),
			removeme->len,
			group_get_readable_name(g));

		group_remove_articles (g, removeme);
	}

	/* cleanup */
	g_ptr_array_free (tmp, TRUE);
	g_ptr_array_free (removeme, TRUE);
	debug_exit ("group_expire_articles_not_in_range");
}


/***
****
***/

static GStaticMutex article_refcount_mutex = G_STATIC_MUTEX_INIT;

static gboolean 
group_ref_articles_if_loaded (Group * group, StatusItem * status)               
{                                                                               
        gboolean retval = FALSE;                                                
        debug_enter ("group_ref_articles_if_loaded");                           

        /* sanity clause */                                                     
        g_return_val_if_fail (group_is_valid(group), FALSE);                    
        g_return_val_if_fail (group->_articles_refcount>=0, FALSE);             

        /* ref if reffed */                                                     
        g_static_mutex_lock (&article_refcount_mutex);                          

        if (group->_articles_refcount > 1) {                                    
                ++group->_articles_refcount;                                    
                retval = TRUE;                                                  
        }                                                                       
        g_static_mutex_unlock (&article_refcount_mutex);                        

        debug_exit ("group_ref_articles_if_loaded");                            

        return retval;                                                          
}                                                                               

void
group_ref_articles (Group * group, StatusItem * status)
{

	debug_enter ("group_ref_articles");

	/* sanity clause */
	g_return_if_fail (group_is_valid(group));
	g_return_if_fail (group->_articles_refcount >= 0);

	g_static_mutex_lock (&article_refcount_mutex);
	if (++group->_articles_refcount == 1)
	{
		g_mutex_lock (group->article_mutex);
		file_headers_load (group, status);
		group_set_article_qty (group, g_hash_table_size(group->_articles));
		g_mutex_unlock (group->article_mutex);
	}
	g_static_mutex_unlock (&article_refcount_mutex);

	debug_exit ("group_ref_articles");
}

void
group_unref_articles (Group * group, StatusItem * status)
{
	debug_enter ("group_unref_articles");

	/* sanity clause */
	g_return_if_fail (group_is_valid(group));
	g_return_if_fail (group->_articles_refcount > 0);

	/* unref */
	g_static_mutex_lock (&article_refcount_mutex);
	if (group->_articles_refcount == 1)
	{
		g_mutex_lock (group->article_mutex);
		file_headers_save_noref (group, NULL);
		server_save_grouplist_if_dirty (group->server, NULL);

		if (group->_articles!=NULL) {
		       	g_hash_table_foreach (group->_articles, destroy_articles_ghfunc, NULL);
			g_hash_table_destroy (group->_articles);
			group->_articles = NULL;
		}
		if (group->_deleted_articles != NULL) {
			pan_g_ptr_array_foreach (group->_deleted_articles, (GFunc)article_destructor, NULL);
			g_ptr_array_free (group->_deleted_articles, TRUE);
			group->_deleted_articles = NULL;
		}
		if (group->_article_chunk != NULL) {
			memchunk_destroy (group->_article_chunk);
			group->_article_chunk = NULL;
		}

		/* free the memory used by those articles */
		replace_gstr (&group->_one_big_chunk, NULL);
		if (group->chunk != NULL) {
			g_string_chunk_free (group->chunk);
			group->chunk = NULL;
		}
		g_mutex_unlock (group->article_mutex);
	}
	--group->_articles_refcount;
	g_static_mutex_unlock (&article_refcount_mutex);

	/* cleanup */
	debug_exit ("group_unref_articles");
}

/**
***
**/


void
group_set_readable_name (Group         * group,
	                 const gchar   * readable_name)
{
	g_return_if_fail (group_is_valid(group));

	replace_gstr (&group->readable_name, g_strdup(readable_name));

	group_set_dirty (group);
	fire_groups_changed (&group, 1);
}


const gchar*
group_get_readable_name (const Group * group)
{
	g_return_val_if_fail (group_is_valid(group), _("(No Group)"));

	return is_nonempty_string (group->readable_name)
		? group->readable_name
		: group->name;
}

/**
***
**/

const gchar*
group_chunk_string (Group         * group,
                    const gchar   * chunkme,
                    gboolean        share)
{
	gchar * retval = NULL;

	g_return_val_if_fail (group_is_valid(group), NULL);

	if (chunkme != NULL)
	{
		if (group->chunk == NULL)
			group->chunk = g_string_chunk_new (81920);

		if (share)
			retval = g_string_chunk_insert_const (group->chunk, chunkme);
		else
			retval = g_string_chunk_insert (group->chunk, chunkme);
	}

	return retval;
}


/***
****
****
****
***/

Newsrc*
group_get_newsrc (Group * group)
{
	g_return_val_if_fail (group_is_valid(group), NULL);

	if (group->_newsrc == NULL)
		group->_newsrc = newsrc_new (NULL, group->article_low, group->article_high);

	return group->_newsrc;
}

gboolean
group_has_newsrc (const Group * group)
{
	g_return_val_if_fail (group_is_valid(group), FALSE);

	return group->_newsrc != NULL;
}

/**
 * When crossposts are deleted, we don't want to load in all the groups
 * containing a crosspost just to delete the article and save the whole
 * group back to disk again.  Likewise the crosspost may not even have been
 * downloaded yet, which would thwart the deletion.
 *
 * So the solution Pan uses is to have a newsrc-like range of article
 * numbers of these deleted articles.  When a client tries to add articles
 * to the group, group_add_articles() checks them against the deleted list
 * before allowing them to be added.
 * 
 * Also important is that we update the deleted list's bounds whenever
 * the group's own low/high numbers are updated, in group_set_article_range(),
 * so that the deleted article list doesn't get too long when saved to disk.
 */
Newsrc*
group_get_purged (Group * group)
{
	g_return_val_if_fail (group_is_valid(group), NULL);

	if (group->_purged == NULL)
		group->_purged = newsrc_new (NULL, group->article_low, group->article_high);

	return group->_purged;
}

static gint
compare_ppA_to_ppA_by_number (const void * va, const void * vb)
{
	const Article * a = *(const Article **)va;
	const Article * b = *(const Article **)vb;
	if (a->number < b->number) return -1;
	if (a->number > b->number) return 1;
	return 0;
}

void
group_remove_crossposts (Group     * g,
		         gulong    * numbers,
                         gint        number_qty)
{
	/* sanity clause */
	g_return_if_fail (group_is_valid(g));
	g_return_if_fail (numbers!=NULL);
	g_return_if_fail (number_qty>0);

	/* remove articles if they're already present */
	if (group_ref_articles_if_loaded (g, NULL))
	{
		gint i;
		GPtrArray * articles;
		GPtrArray * remove;

		/* get a list of articles sorted by number for fast lookup */
		remove = g_ptr_array_new ();
		pan_g_ptr_array_reserve (remove, number_qty);
		articles = group_get_article_array (g);
		qsort (articles->pdata, articles->len, sizeof(Article*), compare_ppA_to_ppA_by_number);

		/* for each number to remove, if we've got it, add it to the "remove" ptr array */
		for (i=0; i<number_qty; ++i)
		{
			Article tmp;
			gint index;
			gboolean exact_match = FALSE;

			tmp.number = numbers[i];
			index = lower_bound (&tmp,
			                     articles->pdata,
			                     articles->len,
			                     sizeof(Article*),
			                     compare_ppA_to_ppA_by_number,
			                     &exact_match);

			if (exact_match) {
				Article * a = ARTICLE(g_ptr_array_index(articles,index));
				g_ptr_array_add (remove, a);
			}
		}

		if (remove->len != 0)
			group_remove_articles (g, remove);

		g_ptr_array_free (remove, TRUE);
		g_ptr_array_free (articles, TRUE);
		group_unref_articles (g, NULL);
	}

	/* flag articles for deletion in case they get downloaded later */
	if (1)
	{
		guint i;
		Newsrc * n;
		
		n = group_get_purged (g);
		for (i=0; i<number_qty; ++i) {
			newsrc_mark_article (n, numbers[i], TRUE);
		}
	}
}

struct _Article*
group_alloc_new_article (Group * group)
{
	Article * a;
	g_return_val_if_fail (group_is_valid(group), NULL);

	if (group->_article_chunk == NULL)
		group->_article_chunk = memchunk_new (sizeof(Article), 200, FALSE);

	a = (Article*) memchunk_alloc (group->_article_chunk);
	a->group = group;
	return a;
}

GHashTable*
group_get_articles (Group * group)
{
	g_return_val_if_fail (group_is_valid(group), NULL);

	/* FIXME: shouldn't this load the articles ??? */
	if (group->_articles == NULL)
		group->_articles = g_hash_table_new (g_str_hash, g_str_equal);

	return group->_articles;
}

GPtrArray*
group_get_article_array (Group * group)
{
	GHashTable * hash;
	GPtrArray * retval;

	/* sanity clause */
	retval = g_ptr_array_new ();
	g_return_val_if_fail (group_is_valid(group), retval);

	/* populate retval */
	hash = group_get_articles (group);
	pan_hash_to_ptr_array (hash, retval);
	pan_warn_if_fail (articles_are_valid ((const Article**)retval->pdata, retval->len));

	return retval;
}

/***
****
****  EVENTS
****
***/
 
/**
***  Groups Removed
**/
 
PanCallback*
group_get_groups_changed_callback (void)
{
	static PanCallback * cb = NULL;
	if (cb==NULL) cb = pan_callback_new ();
	return cb;
}

static void
fire_groups_changed (Group ** g, gint qty)
{
	g_return_if_fail (g!=NULL);

	pan_callback_call (group_get_groups_changed_callback(), g, GINT_TO_POINTER(qty));
}

PanCallback*
group_get_articles_added_callback (void)
{
	static PanCallback * cb = NULL;
	if (cb==NULL) cb = pan_callback_new ();
	return cb;
}

PanCallback*
group_get_articles_removed_callback (void)
{
	static PanCallback * cb = NULL;
	if (cb==NULL) cb = pan_callback_new ();
	return cb;
}
